#!/bin/sh

# Commands to run this script:
# fakeroot ./makedeb.sh

# This script is to be used to generate the Debian package for
# CCT. It assumes that files required to create the Debian
# package is under Linux/Debian of CCT source directory.

# Enters the DIR where this script is in.
REL_DIR=`dirname $0`
case "$1" in
    cct)
    echo "Building cct package for Debian"
    ;;
    cct-fonts)
    echo "Building cct-fonts package for Debian"
    ;;
    *)
    echo "Usage: ./makedeb.sh {cct|cct-fonts}"
    exit 1;
esac

TARGET="$1"

cd ${REL_DIR}

# Goes back to the source directory
cd ../..
CCT_SRC_DIR=`pwd`

if [ "$TARGET" = "cct" ]
then
# Make sure the libkpathsea-dev is installed.
    #if ! dpkg -s libkpathsea-dev; then
    if ! test -r /usr/include/kpathsea/kpathsea.h; then
	cat 1>&2 <<END
Cannot find "/usr/include/kpathsea/kpathsea.h".
Please install the 'libkpathsea-dev' package first.
It is required to compile CCT.
END
	exit 1
    fi
# Builds the package and then installs it into a temp directory
# which is Linux/Debian/cct of CCT source directory.
    if env LANG=C gcc -v 2>&1 | grep -q 'gcc version 4'; then
	make EXTRA_CFLAGS=-Wno-pointer-sign all
    else
	make all
    fi
    make install_root=${CCT_SRC_DIR}/Linux/Debian/cct prefix=/usr install
    cd Linux/Debian
# Generates md5sums
    if test -f cct/DEBIAN/md5sums; then
        rm -f cct/DEBIAN/md5sums
    fi
    find cct/usr -type f -exec md5sum {} \;>>cct/DEBIAN/md5sums
# Archives the package
    dpkg -b cct .
    echo 1>&2 "The package file is saved in '`pwd`'"
    exit 0
fi

if [ "$TARGET" = "cct-fonts" ]
then
    cd Linux/Debian
# Generates md5sums
    if test -f cct-fonts/DEBIAN/md5sums; then
        rm -f cct-fonts/DEBIAN/md5sums
    fi
    find cct-fonts/usr -type f -exec md5sum {} \;>>cct-fonts/DEBIAN/md5sums
# Archives the package
    dpkg -b cct-fonts .
    echo 1>&2 "The package file is saved in '`pwd`'"
    exit 0
fi

exit 0
